@@ -1,19 +1,19 @@
+#!perl
+
 use strict;
 use warnings;
-use feature qw(say);
+use feature qw(say state);
 
 package App::GitHubPullRequest;
-{
-  $App::GitHubPullRequest::VERSION = '0.3.1';
-}
-
+$App::GitHubPullRequest::VERSION = '0.4.0';
 # ABSTRACT: Command-line tool to query GitHub pull requests
 
 use JSON qw(decode_json encode_json);
 use Carp qw(croak);
 use Encode qw(encode_utf8);
+use File::Spec;
 
-sub DEBUG;
+use constant DEBUG => $ENV{'GIT_PR_DEBUG'} || 0;
 
 
 sub new {
@@ -44,7 +44,8 @@ Where command is one of these:
   patch <number>    Fetch a properly formatted patch for specified pull request
   checkout <number> Create tracking branch for specified pull request
 
-  login [<user>] [<password>] Login to GitHub and receive an access token
+  login [<user>] [<password>] [<two-factor-auth-token>]
+                              Login to GitHub and receive an access token
   comment <number> [<text>]   Create a comment on the specified pull request
   close <number>              Close the specified pull request
   open <number>               Reopen the specified pull request
@@ -56,13 +57,15 @@ EOM
 
 sub list {
     my ($self, $state) = @_;
-    my $prs = $self->_fetch_all($state);
-    say ucfirst($prs->{'state'}) . " pull requests for '" . $prs->{"repo"} . "':";
-    unless ( $prs->{'pull_requests'} and @{ $prs->{'pull_requests'} } ) {
+    $state ||= 'open';
+    my $remote_repo = _find_github_remote();
+    my $prs = _api_read("/repos/$remote_repo/pulls?state=$state");
+    say ucfirst($state) . " pull requests for '$remote_repo':";
+    unless ( @$prs ) {
         say "No pull requests found.";
         return 0;
     }
-    foreach my $pr ( @{ $prs->{"pull_requests"} } ) {
+    foreach my $pr ( @$prs ) {
         my $number = $pr->{"number"};
         my $title = encode_utf8( $pr->{"title"} );
         my $date = $pr->{"updated_at"} || $pr->{'created_at'};
@@ -89,7 +92,7 @@ sub show {
         say "Number:  $number";
         say "\n$body\n" if $body;
     }
-    my $comments = $self->_fetch_comments($pr);
+    my $comments = _api_read( $pr->{'comments_url'} );
     foreach my $comment (@$comments) {
         my $user = $comment->{'user'}->{'login'};
         my $date = $comment->{'updated_at'} || $comment->{'created_at'};
@@ -106,7 +109,9 @@ sub show {
 sub patch {
     my ($self, $number) = @_;
     die("Please specify a pull request number.\n") unless $number;
-    my $patch = $self->_fetch_patch($number);
+    my $patch = _get_url(
+        $self->_fetch_one($number)->{'patch_url'}
+    );
     die("Unable to fetch patch for pull request $number.\n")
         unless defined $patch;
     print $patch;
@@ -116,10 +121,9 @@ sub patch {
 
 sub checkout {
     my ($self, $number) = @_;
+    $number =~ s{[^\d]}{}g if defined $number;
     die("Please specify a pull request number.\n") unless $number;
     my $pr = $self->_fetch_one($number);
-    die("Unable to fetch pull request $number.\n")
-        unless defined $pr;
 
     # Get required contributor branch info
     my $head_repo   = $pr->{'head'}->{'repo'}->{'git_url'};
@@ -226,7 +230,7 @@ sub comment {
         system($editor, $filename);
         # Fetch text just edited (if any)
         if ( -r $filename ) {
-            CORE::open my $fh, "<", $filename or die "Can't open $filename: $!";
+            CORE::open my $fh, "<:encoding(UTF-8)", $filename or die "Can't open $filename: $!";
             $text = join "", <$fh>;
             CORE::close $fh;
         }
@@ -236,11 +240,11 @@ sub comment {
             die("Your comment is empty. Command aborted.\n");
         }
     }
-    my $url = "https://api.github.com/repos/$remote_repo/issues/$number/comments";
-    my $mimetype = 'application/json';
-    my $data = encode_json({ "body" => $text });
     my $comment = eval {
-        decode_json( _post_url($url, $mimetype, $data) );
+        _api_create(
+            "/repos/$remote_repo/issues/$number/comments",
+            { "body" => $text },
+        );
     };
     die($@ . ( defined $filename ? "Comment text saved in '$filename'. Please remove it manually." : "" ) ."\n")
         if $@; # most likely network error
@@ -250,9 +254,9 @@ sub comment {
 
     # Remove temporary file if everything went well
     if ( defined $filename and -e $filename ) {
-        unlink $filename;
+        my $count = unlink $filename;
         warn("Unable to remove temporary file $filename: $!\n")
-            if $!;
+            unless $count;
     }
 
     return 0;
@@ -260,28 +264,39 @@ sub comment {
 
 
 sub login {
-    my ($self, $user, $password) = @_;
+    my ($self, $user, $password, $two_factor_token) = @_;
+
+    # Try to fetch user/password from git config (or prompt)
     $user     ||= _qx('git', "config github.user")     || _prompt('GitHub username');
     $password ||= _qx('git', "config github.password") || _prompt('GitHub password', 'hidden');
     die("Please specify a user name.\n") unless $user;
     die("Please specify a password.\n")  unless $password;
-    my $url = "https://api.github.com/authorizations";
-    my $mimetype = 'application/json';
-    my $data = encode_json({
-        "scopes"   => [qw( public_repo repo )],
-        "note"     => __PACKAGE__,
-        "note_url" => 'https://metacpan/module/' . __PACKAGE__,
-    });
-    my $auth = decode_json( _post_url($url, $mimetype, $data, $user, $password) );
+    # Prompt for two-factor auth token
+    $two_factor_token ||= _prompt('GitHub two-factor authentication token (if any)');
+
+    # Perform authentication
+    my $auth = _api_create(
+        "/authorizations",
+        {
+            "scopes"   => [qw( public_repo repo )],
+            "note"     => __PACKAGE__,
+            "note_url" => 'https://metacpan/module/' . __PACKAGE__,
+        },
+        $user,
+        $password,
+        $two_factor_token,
+    );
     die("Unable to authenticate with GitHub.\n")
         unless defined $auth;
     my $token = $auth->{'token'};
     die("Authentication data does not include a token.\n")
         unless $token;
+
+    # Store successful authorization token
     my ($content, $rc) = _run_ext(qw(git config --global github.pr-token), $token);
     die("git config returned message '$content' and code $rc when trying to store your token.\n")
         if $rc != 0;
-    say "Access token stored successfully. Go to https://github.com/settings/applications to revoke access.";
+    say "Access token stored successfully. Go to https://github.com/settings/tokens to revoke access.";
     return 0;
 }
 
@@ -290,92 +305,99 @@ sub _state {
     croak("Please specify a pull request number") unless $number;
     croak("Please specify a pull request state") unless $state;
     my $remote_repo = _find_github_remote();
-    my $url = "https://api.github.com/repos/$remote_repo/pulls/$number";
-    my $mimetype = 'application/json';
-    my $data = encode_json({ "state" => $state });
-    my $pr = decode_json( _patch_url($url, $mimetype, $data) );
-    return $pr;
-}
-
-sub _fetch_comments {
-    my ($self, $pr) = @_;
-    croak("Please specify a pull request") unless $pr;
-    my $comments_url = $pr->{'comments_url'};
-    my $comments = decode_json( _get_url($comments_url) );
-    return $comments;
-}
-
-sub _fetch_patch {
-    my ($self, $number) = @_;
-    my $patch_url = $self->_fetch_one($number)->{'patch_url'};
-    return _get_url($patch_url);
+    return _api_update(
+        "/repos/$remote_repo/pulls/$number",
+        { "state" => $state },
+    );
 }
 
 sub _fetch_one {
     my ($self, $number) = @_;
     my $remote_repo = _find_github_remote();
-    my $pr_url = "https://api.github.com/repos/$remote_repo/pulls/$number";
-    my $pr = decode_json( _get_url($pr_url) );
-    return $pr;
-}
-
-sub _fetch_all {
-    my ($self, $state) = @_;
-    $state ||= 'open';
-    my $remote_repo = _find_github_remote();
-    my $pulls_url = "https://api.github.com/repos/$remote_repo/pulls?state=$state";
-    my $pull_requests = decode_json( _get_url($pulls_url) );
-    return {
-        "repo"           => $remote_repo,
-        "state"          => $state,
-        "pull_requests"  => $pull_requests,
-    };
+    return _api_read("/repos/$remote_repo/pulls/$number");
 }
 
 sub _find_github_remote {
     # Parse lines from git and use first found github repo
-    my $repo;
+    my @repos;
+    my $repo_map = {};
     foreach my $line ( _qx('git', 'remote -v') ) {
         my ($remote, $url, $type) = split /\s+/, $line;
         next unless $type eq '(fetch)'; # only consider fetch remotes
         next unless $url =~ m/github\.com/; # only consider remotes to github
         if ( $url =~ m{github.com[:/](.+?)(?:\.git)?$} ) {
-            $repo = $1;
-            last;
+            push @repos, $1;
+            $repo_map->{$1} = $remote;
         }
     }
 
     # Allow override for testing
-    $repo = $ENV{"GITHUB_REPO"} if $ENV{'GITHUB_REPO'};
-    die("No valid GitHub remote repo found.\n")
-        unless $repo;
+    @repos = ( $ENV{"GITHUB_REPO"} ) if $ENV{'GITHUB_REPO'};
 
-    # Fetch repo information
-    my $repo_url = "https://api.github.com/repos/$repo";
-    my $repo_info = decode_json( _get_url( $repo_url ) );
-    die("Unable to fetch repo information for $repo_url.\n")
-        unless $repo_info;
+    die("No valid GitHub repos found.\n")
+        unless @repos;
 
-    # Return the parent repo if repo is a fork
-    return $repo_info->{'parent'}->{'full_name'}
-        if $repo_info->{'fork'};
+    # Try each repo found in turn
+    while ( my $repo = shift @repos ) {
+        print "Fetching GitHub repo: $repo\n"
+            if DEBUG;
+        # Fetch repo information
+        my ($repo_info, $code) = _api_read("/repos/$repo", 'return_on_error');
 
-    # Not a fork, use this repo
-    return $repo;
+        # Skip repo which is not found
+        if ( $code eq 404 ) {
+            print STDERR "WARNING: Skipping invalid GitHub repo: $repo\n";
+            if ( $repo_map->{$repo} ) {
+                print STDERR "WARNING:\n";
+                print STDERR "WARNING: Remove invalid git remote with command:\n";
+                print STDERR "WARNING:   git remote remove $repo_map->{$repo}\n";
+                print STDERR "WARNING:\n";
+            }
+            next;
+        }
+
+        # Return the parent repo if repo is a fork
+        return $repo_info->{'parent'}->{'full_name'}
+            if $repo_info->{'fork'};
+
+        # Not a fork, use this repo
+        return $repo;
+    }
+
+    die("No valid GitHub repo found. List of remotes exhausted.\n");
 }
 
 # Ask the user for some information
+# Disable local echo if $hide_echo is true (for passwords)
 sub _prompt {
     my ($label, $hide_echo) = @_;
+    _echo('off') if $hide_echo;
     print "$label: " if defined $label;
-    _require_binary('stty') if $hide_echo;
-    system("stty -echo") if $hide_echo;
     my $input = scalar <STDIN>;
-    system("stty echo") if $hide_echo;
     chomp $input;
+    print "\n";
+    _echo('on') if $hide_echo;
     return $input;
 }
 
+# Turn local echo on or off (Unix only for now)
+sub _echo {
+    my ($state) = @_;
+    croak("Please specify an echo state of on or off") unless $state;
+
+    # Test if we're on Unix (snatched from Platform::Unix)
+    my $is_unix = $^O =~ /^(Linux|.*BSD.*|.*UNIX.*|Darwin|Solaris|SunOS|Haiku|Next|dec_osf|svr4|sco_sv|unicos.*|.*x)$/i;
+
+    if ( $is_unix ) {
+        return _run_ext(qw{stty -echo}) if $state eq 'off';
+        return _run_ext(qw{stty echo})  if $state eq 'on';
+    }
+
+    # Don't know how to turn local echo on or off on other platforms.
+    # If you happen to know how, please send a pull request with a fix
+    return;
+}
+
 # Generate a random temporary filename with the given prefix
 sub _tmpfile {
     my ($prefix) = @_;
@@ -394,6 +416,7 @@ sub _qx {
     my ($cmd, @rest) = @_;
     _require_binary($cmd);
     $cmd .= " " . join(" ", @rest) if @rest;
+    warn("_qx: $cmd\n") if DEBUG;
     return map { chomp; $_ } qx{$cmd}
         if wantarray;
     my $content = qx{$cmd};
@@ -407,7 +430,7 @@ sub _run_ext {
     croak("Please specify a command line") unless @_;
     my $cmd = join(" ", @_);
     my $prg = $_[0];
-    warn("$cmd\n") if DEBUG;
+    warn("_run_ext: $cmd\n") if DEBUG;
     _require_binary($prg);
     CORE::open my $fh, "-|", @_ or die("Can't run command '$cmd': $!");
     my $stdout = join("", <$fh>);
@@ -421,24 +444,95 @@ sub _run_ext {
 sub _require_binary {
     my ($bin) = @_;
     croak("Please specify program to require") unless $bin;
-    system("which $bin >/dev/null");
-    return 1 if $? >> 8 == 0; # exit code is 0
+    state %cache;
+    unless (exists $cache{$bin}) {
+        # Cache miss, let's see if we can find that program
+        warn("Checking if '$bin' exists in path\n") if DEBUG;
+        # Figure out path elements in a cross-platform way and go through
+        # each and check if it has the executable program in it
+        my @path = File::Spec->path();
+        while (my $dir = shift @path) {
+            my $file = File::Spec->catfile($dir, $bin);
+            warn("Looking for executable bit on '$file'\n") if DEBUG;
+            ## no critic (ValuesAndExpressions::ProhibitCommaSeparatedStatements)
+            $cache{$bin} = $file, last if -x $file;
+            ## use critic
+            $cache{$bin} = 0;
+        }
+        warn("'$bin' was found at '$cache{$bin}'\n") if DEBUG and $cache{$bin};
+    }
+    return 1 if $cache{$bin};
     die("You need the program '$bin' in your path to use this feature.\n");
 }
 
+# Return the base GitHub API URL as mentioned
+# on http://developer.github.com/v3/
+sub _api_url {
+    my ($url) = @_;
+    my $prefix = 'https://api.github.com/';
+    # If no URL specified, just return the API URL
+    return $prefix unless defined $url;
+    # If URL already looks like a GitHub API URL, do nothing
+    return $url if _is_api_url($url);
+    # Create an API URL out of a partial URL as
+    $url =~ s{^/*}{}; # Remove initial slashes, if any
+    return $prefix . $url;
+}
 
-sub DEBUG {
-    return $ENV{'GIT_PR_DEBUG'} || 0;
+# Check if a URL is a GitHub API URL
+sub _is_api_url {
+    my ($url) = @_;
+    croak("Please specify a URL to verify") unless $url;
+    my $prefix = _api_url();
+    return 1 if index($url, $prefix) == 0;
+    return 0;
+}
+
+# Perform an API GET request
+sub _api_read {
+    my ($url, $return_on_error) = @_;
+    my ($response, $code) = _get_url(
+        _api_url($url),
+        $return_on_error,
+    );
+    return decode_json($response), $code
+        if $return_on_error;
+    return decode_json($response);
+}
+
+# Perform an API POST request
+sub _api_create {
+    my ($url, $data, @rest) = @_;
+    return decode_json(
+        _post_url(
+            _api_url($url),
+            'application/json',
+            encode_json($data),
+            @rest,
+        )
+    );
+}
+
+# Perform an API PATCH request
+sub _api_update {
+    my ($url, $data) = @_;
+    return decode_json(
+        _patch_url(
+            _api_url($url),
+            'application/json',
+            encode_json($data),
+        )
+    );
 }
 
 # Perform HTTP GET
 sub _get_url {
-    my ($url) = @_;
+    my ($url, $return_on_error) = @_;
     croak("Please specify a URL") unless $url;
 
     # See if we should use credentials
     my @credentials;
-    if ( $url =~ m{^https://api.github.com/} ) {
+    if ( _is_api_url($url) ) {
         my $token = _qx('git', 'config github.pr-token');
         @credentials = ( '-H', "Authorization: token $token" ) if $token;
     }
@@ -446,6 +540,7 @@ sub _get_url {
     # Send request
     my ($content, $rc) = _run_ext(
         'curl',
+        '-L',                            # follow redirects
         '-s',                            # be silent
         '-w', '%{http_code}',            # include HTTP status code at end of stdout
         @credentials,                    # Logon credentials, if any
@@ -454,6 +549,9 @@ sub _get_url {
     die("curl failed to fetch $url with code $rc.\n") if $rc != 0;
 
     my $code = substr($content, -3, 3, '');
+
+    return $content, $code if $return_on_error;
+
     if ( $code >= 400 ) {
         die("Fetching URL $url failed with code $code:\n$content");
     }
@@ -470,7 +568,7 @@ sub _patch_url {
 
     # See if we should use credentials
     my @credentials;
-    if ( $url =~ m{^https://api.github.com/} ) {
+    if ( _is_api_url($url) ) {
         my $token = _qx('git', 'config github.pr-token');
         die("You must login before you can modify information.\n")
             unless $token;
@@ -506,19 +604,21 @@ sub _patch_url {
 
 # Perform HTTP POST
 sub _post_url {
-    my ($url, $mimetype, $data, $user, $password) = @_;
+    my ($url, $mimetype, $data, $user, $password, $two_factor_token) = @_;
     croak("Please specify a URL") unless $url;
     croak("Please specify a mimetype") unless $mimetype;
     croak("Please specify some data") unless $data;
 
     # See if we should use credentials
     my @credentials;
-    if ( $url =~ m{^https://api.github.com/} ) {
+    if ( _is_api_url($url) ) {
         my $token = _qx('git', 'config github.pr-token');
         die("You must login before you can modify information.\n")
             unless $token or ( $user and $password );
         if ( $user and $password ) {
             @credentials = ( '-u', "$user:$password" );
+            push @credentials, '-H', "X-GitHub-OTP: $two_factor_token"
+                if $two_factor_token;
         }
         else {
             @credentials = ( '-H', "Authorization: token $token" ) if $token;
@@ -552,7 +652,7 @@ __END__
 
 =pod
 
-=encoding utf-8
+=encoding UTF-8
 
 =head1 NAME
 
@@ -560,7 +660,7 @@ App::GitHubPullRequest - Command-line tool to query GitHub pull requests
 
 =head1 VERSION
 
-version 0.3.1
+version 0.4.0
 
 =head1 SYNOPSIS
 
@@ -651,12 +751,14 @@ edit the text.  If it has not been set, it will try to use the L<editor(1)>
 external program.  This is usually a symlink set up by your operating system
 to the most recently installed text editor.
 
-=head2 login [<user>] [<password>]
+The text must be encoded using UTF-8.
+
+=head2 login [<user>] [<password>] [<2fa-token>]
 
 Logs you in to GitHub and creates a new access token used instead of your
-password.  If you don't specify either of the options, they are looked up in
-your git config github section.  If none of those are found, you'll be
-prompted for them.
+password and two-factor authentication token.  If you don't specify either
+of the options, they are looked up in your git config github section.  If
+none of those are found, you'll be prompted for them.
 
 =head1 METHODS
 
@@ -669,17 +771,11 @@ Constructor. Takes no parameters.
 Calls any of the other listed public methods with specified arguments. This
 is usually called automatically when you invoke L<git-pr>.
 
-=head1 FUNCTIONS
-
-=head2 DEBUG
-
-Returns true if we're in debug mode.
+=head1 DEBUGGING
 
 Set the environment variable GIT_PR_DEBUG to a non-zero value to see more
 details, like each API command being executed.
 
-=head1 DEBUGGING
-
 If you want to interact with another GitHub repo than the one in your
 current directory, set the environment variable GITHUB_REPO to the name of
 the repo in question. Example:
@@ -831,7 +927,7 @@ The code is open to the world, and available for you to hack on. Please feel fre
 with it, or whatever. If you want to contribute patches, please send me a diff or prod me to pull
 from your repository :)
 
-L<http://github.com/robinsmidsrod/App-GitHubPullRequest>
+L<https://github.com/robinsmidsrod/App-GitHubPullRequest>
 
   git clone git://github.com/robinsmidsrod/App-GitHubPullRequest.git
 
@@ -841,7 +937,7 @@ Robin Smidsrød <robin@smidsrod.no>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2013 by Robin Smidsrød.
+This software is copyright (c) 2016 by Robin Smidsrød.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
